#include "buffer.h"

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include "md5.h"

#include <new>

Buffer::Buffer() {
	data = NULL;
	length = 0;
	blocksize = 0;
	pos = 0;
	error = false;
}
Buffer::~Buffer() {
	free(data);
}
void Buffer::Clear() {
	free(data);
	data = NULL;
	length = 0;
	blocksize = 0;
	pos = 0;
	error = false;
}
void Buffer::SetPos(unsigned long newpos) {
	pos = (newpos>length)? length : newpos;
}
void Buffer::SetLength(unsigned long newlength) {
	if(newlength>=blocksize || newlength+64<(blocksize>>1)) {
		unsigned long i = newlength | (128-1);
		i |= i>>1;
		i |= i>>2;
		i |= i>>4;
		i |= i>>8;
		i |= i>>16;
		++i;
		if(i!=blocksize) {
			if(length==0 && i>blocksize) {
				free(data);
				blocksize = 0;
				data = NULL;
			}
			void *temp = realloc(data, i);
			if(temp==NULL) {
				throw std::bad_alloc();
			}
			blocksize = i;
			data = (char*)(temp);
		}
	}
	length = newlength;
	if(pos>length) pos = length;
	data[length] = '\0';
}
void Buffer::Swap(Buffer* other) {
	{ char *temp = data; data = other->data; other->data = temp; }
	{ unsigned long temp = length; length = other->length; other->length = temp; }
	{ unsigned long temp = blocksize; blocksize = other->blocksize; other->blocksize = temp; }
	{ unsigned long temp = pos; pos = other->pos; other->pos = temp; }
	{ bool temp = error; error = other->error; other->error = temp; }
}

bool Buffer::ReadFromFile(const char* filename) {
	Clear();
	FILE *f = fopen(filename, "rb");
	if(f==NULL) return false;
	try {
		fseek(f, 0, SEEK_END);
		unsigned long size = ftell(f);
		fseek(f, 0, SEEK_SET);
		if(size==(unsigned long)(-1)) {
			fclose(f);
			return false;
		}
		SetLength(size);
		if(fread(data, 1, size, f)!=size) {
			Clear();
			fclose(f);
			return false;
		}
	}
	catch(...) {
		fclose(f);
		throw;
	}
	fclose(f);
	return true;
}
bool Buffer::WriteToFile(const char* filename) {
	FILE *f;
	f = fopen(filename, "wb");
	if(f==NULL) return false;
	if(fwrite(data, 1, length, f)!=length) {
		fclose(f);
		return false;
	}
	fclose(f);
	return true;
}

const char* Buffer::ReadString() {
	if(error) return "";
	for(unsigned long i = pos; i<length; ++i) {
		if(data[i]=='\0') {
			const char *str = data+pos;
			pos = i+1;
			return str;
		}
	}
	error = true;
	return "";
}
void Buffer::WriteString(const char* string) {
	unsigned long len = strlen(string)+1;
	if(len!=0) {
		unsigned long p = length;
		SetLength(length+len);
		memcpy(data+p, string, len);
	}
}

void Buffer::ReadData(char* ptr, unsigned long len) {
	if(error) {
		memset(ptr, 0, len);
		return;
	}
	if(pos+len>length) {
		error = true;
		memset(ptr, 0, len);
		return;
	}
	if(len!=0) {
		memcpy(ptr, data+pos, len);
		pos += len;
	}
}
void Buffer::WriteData(const char* ptr, unsigned long len) {
	if(len!=0) {
		unsigned long p = length;
		SetLength(pos+len);
		memcpy(data+p, ptr, len);
	}
}

void Buffer::ReadHex(char* ptr, unsigned long len) {
	if(error) {
		memset(ptr, '0', len*2);
		return;
	}
	if(pos+len>length) {
		error = true;
		memset(ptr, '0', len*2);
		return;
	}
	if(len!=0) {
		bintohex(data+pos, ptr, len);
		pos += len;
	}
}
void Buffer::WriteHex(const char* ptr, unsigned long len) {
	if(len!=0) {
		unsigned long p = length;
		SetLength(pos+len);
		hextobin(ptr, data+p, len);
	}
}

void Buffer::WriteBuffer(const Buffer* b) {
	unsigned long len = b->length; // in case b = this
	if(len!=0) {
		unsigned long p = length;
		SetLength(length+len);
		memcpy(data+p, b->data, len);
	}
}
void Buffer::WriteBufferPart(const Buffer* b, unsigned long _pos, unsigned long _len) {
	if(_pos>b->length) _pos = b->length;
	if(_len>b->length-_pos) _len = b->length-_pos;
	if(_len!=0) {
		unsigned long p = length;
		SetLength(length+_len);
		memcpy(data+p, b->data+_pos, _len);
	}
}

