#include "HttpRequest.h"

#include <cassert>
#include <cstring>

struct HttpRequest_UrlEncodeTable {
	unsigned char table[256];
	HttpRequest_UrlEncodeTable() {
		memset(table, 0, 256);
		for(const char *p = "%$.+!*'(),;/?:@=&#"; *p!='\0'; ++p) { table[(unsigned char)(*p)] = 1; }
		for(char c = '0'; c<='9'; ++c) { table[(unsigned char)(c)] = 2; }
		for(char c = 'a'; c<='z'; ++c) { table[(unsigned char)(c)] = 2; }
		for(char c = 'A'; c<='Z'; ++c) { table[(unsigned char)(c)] = 2; }
	}
} httprequest_urlencodetable;

bool HttpRequest::HeaderExists(const char* name) {
	for(unsigned int i = 0; i<headers.size(); ++i) {
		if(_strnicmp(headers[i].name.data(), name, headers[i].name.length())==0) {
			return true;
		}
	}
	return false;
}

HttpRequest::HttpRequest() {
	state = HTTPREQUESTSTATE_NOTCONNECTED;
}

void HttpRequest::SetHeader(const char* name, const char* value, bool replace) {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	if(replace) {
		for(unsigned int i = 0; i<headers.size(); ++i) {
			if(_strnicmp(headers[i].name.data(), name, headers[i].name.length())==0) {
				headers[i].value = value;
				return;
			}
		}
	}
	headers.push_back(Header(name, value));
}
void HttpRequest::RemoveHeader(const char* name) {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	for(unsigned int i = headers.size(); i>0; ) {
		--i;
		if(_strnicmp(headers[i].name.data(), name, headers[i].name.length())==0) {
			headers.erase(headers.begin()+i);
		}
	}
}
void HttpRequest::ClearHeaders() {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	headers.clear();
}

void HttpRequest::SetPostParameter(const char* name, const char* value) {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	for(unsigned int i = 0; i<postparameters.size(); ++i) {
		if(postparameters[i].name.compare(name)==0) {
			postparameters[i].value = value;
			return;
		}
	}
	postparameters.push_back(PostParameter(name, "", value));
}
void HttpRequest::SetPostParameterFile(const char* name, const char* filename, const char* value, unsigned int valuelength) {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	for(unsigned int i = 0; i<postparameters.size(); ++i) {
		if(postparameters[i].name.compare(name)==0) {
			postparameters[i].value.assign(value, valuelength);
			return;
		}
	}
	postparameters.push_back(PostParameter(name, filename, value, valuelength));
}
void HttpRequest::RemovePostParameter(const char* name) {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	for(unsigned int i = 0; i<postparameters.size(); ++i) {
		if(postparameters[i].name.compare(name)==0) {
			postparameters.erase(postparameters.begin()+i);
			return;
		}
	}
}
void HttpRequest::ClearPostParameters() {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	postparameters.clear();
}

void HttpRequest::Connect(const char* _url, bool _post) {
	assert(state==HTTPREQUESTSTATE_NOTCONNECTED);
	
	// get host, port and file
	std::string host, port, file;
	{
		const char *pos_host = strstr(_url, "://");
		if(pos_host==NULL) {
			pos_host = _url;
		} else {
			pos_host += 3;
		}
		const char *pos_port = strchr(pos_host, ':');
		if(pos_port==NULL) {
			port.assign("80");
			const char *pos_file = strchr(pos_host, '/');
			if(pos_file==NULL) {
				host.assign(pos_host);
				file.assign("/");
			} else {
				host.assign(pos_host, pos_file-pos_host);
				file.assign(pos_file);
			}
		} else {
			host.assign(pos_host, pos_port-pos_host);
			const char *pos_file = strchr(pos_port, '/');
			if(pos_file==NULL) {
				port.assign(pos_port+1);
				file.assign("/");
			} else {
				port.assign(pos_port+1, pos_file-pos_port-1);
				file.assign(pos_file);
			}
		}
	}
	file = UrlEncode(file.c_str(), true);
	
	s.Create(host.c_str(), port.c_str());
	/*{
		headers.clear();
		postparameters.clear();
		statuscode = 0;
		state = HTTPREQUESTSTATE_ERROR;
		return;
	}*/
	{
		const char *method = (_post)? "POST " : "GET ";
		const char *httpversion = " HTTP/1.0\r\n";
		s.Write(method, strlen(method));
		s.Write(file.data(), file.length());
		s.Write(httpversion, strlen(httpversion));
	}
	if(!HeaderExists("Host")) {
		headers.push_back(Header("Host", host.c_str()));
	}
	std::string sendmessagebody;
	if(_post) {
		bool use_multipart = false;
		for(unsigned int i = 0; i<postparameters.size(); ++i) {
			if(!postparameters[i].filename.empty()) {
				use_multipart = true;
				break;
			}
		}
		std::string content_type;
		if(use_multipart) {
			srand(clock());
			std::string boundary;
			for( ; ; ) {
				boundary.assign(25, '-');
				for(unsigned int i = 0; i<15; ++i) {
					boundary.push_back('a'+rand()%26);
				}
				for(unsigned int i = 0; i<postparameters.size(); ++i) {
					if(postparameters[i].name.find(boundary)!=std::string::npos
						|| postparameters[i].filename.find(boundary)!=std::string::npos
						|| postparameters[i].value.find(boundary)!=std::string::npos) {
						continue;
					}
				}
				break;
			}
			for(unsigned int i = 0; i<postparameters.size(); ++i) {
				sendmessagebody.append("--");
				sendmessagebody.append(boundary);
				sendmessagebody.append("\r\nContent-Disposition: form-data; name=\"");
				sendmessagebody.append(postparameters[i].name);
				if(!postparameters[i].filename.empty()) {
					sendmessagebody.append("\"; filename=\"");
					sendmessagebody.append(postparameters[i].filename);
				}
				sendmessagebody.append("\"\r\n\r\n");
				sendmessagebody.append(postparameters[i].value);
				sendmessagebody.append("\r\n");
			}
			sendmessagebody.append("--");
			sendmessagebody.append(boundary);
			sendmessagebody.append("--\r\n");
			content_type = "multipart/form-data; boundary=";
			content_type.append(boundary);
		} else {
			for(unsigned int i = 0; i<postparameters.size(); ++i) {
				if(i!=0) sendmessagebody.append("&");
				sendmessagebody.append(postparameters[i].name);
				sendmessagebody.append("=");
				sendmessagebody.append(UrlEncode(postparameters[i].value.c_str()));
			}
			content_type = "application/x-www-form-urlencoded";
		}
		char buffer[21];
		buffer[20] = '\0';
		char *ptr = buffer+20;
		unsigned long s = sendmessagebody.size();
		do {
			*(--ptr) = '0'+s%10;
		} while((s /= 10)!=0);
		SetHeader("Content-Type", content_type.c_str());
		SetHeader("Content-Length", ptr);
	}
	for(unsigned int i = 0; i<headers.size(); ++i) {
		s.Write(headers[i].name.data(), headers[i].name.length());
		s.Write(": ", 2);
		s.Write(headers[i].value.data(), headers[i].value.length());
		s.Write("\r\n", 2);
	}
	s.Write("\r\n", 2);
	s.Write(sendmessagebody.data(), sendmessagebody.length());
	
	headers.clear();
	postparameters.clear();
	statuscode = 0;
	state = HTTPREQUESTSTATE_WAIT_STATUSCODE;
	
}

void HttpRequest::Update() {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	
	if(state==HTTPREQUESTSTATE_ERROR) {
		return;
	}
	if(!s.Update()) {
		return;
	}
	
	if(state==HTTPREQUESTSTATE_WAIT_STATUSCODE) {
		const char *data = s.GetReadData();
		unsigned long len = s.GetReadDataLength();
		unsigned long linelen;
		for(linelen = 0; linelen+1<len; ++linelen) {
			if(data[linelen]=='\r' && data[linelen+1]=='\n') break;
		}
		if(linelen+1<len) {
			// skip http version
			unsigned long i;
			for(i = 0; i<linelen; ++i) {
				if(data[i]==' ') break;
			}
			++i;
			// read status code
			unsigned long j;
			for(j = i; j<linelen; ++j) {
				if(data[j]==' ') break;
			}
			statuscode = 0;
			for(; i<j; ++i) {
				char c = data[i];
				if(c<'0' || c>'9') break;
				statuscode = statuscode*10+(unsigned char)(c-'0');
			}
			s.EraseReadData(linelen+2);
			state = HTTPREQUESTSTATE_WAIT_HEADERS;
		}
	}
	
	if(state==HTTPREQUESTSTATE_WAIT_HEADERS) {
		const char *data = s.GetReadData();
		unsigned long len = s.GetReadDataLength();
		unsigned long headerslen;
		for(headerslen = 0; headerslen+3<len; ++headerslen) {
			if(data[headerslen]=='\r' && data[headerslen+1]=='\n'
				&& data[headerslen+2]=='\r' && data[headerslen+3]=='\n') break;
		}
		if(headerslen+3<len) {
			unsigned long i = 0;
			while(i<headerslen) {
				unsigned long j;
				for(j = i; j+1<headerslen; ++j) {
					if(data[j]=='\r' && data[j+1]=='\n') {
						if(j+2>=headerslen) break;
						if(data[j+2]!=' ' && data[j+2]!='\t') break;
					}
				}
				if(j+1>=headerslen) {
					j = headerslen;
				}
				unsigned long k;
				for(k = i; k<j; ++k) {
					if(data[k]==':') break;
				}
				headers.push_back(Header());
				headers.back().name.assign(data+i, k-i);
				++k;
				unsigned long s = 0;
				bool whitespace = true;
				for(i = k; i<j; ++i) {
					if(data[i]==' ' || data[i]=='\t') {
						if(!whitespace) {
							++s;
						}
						whitespace = true;
					} else {
						++s;
						whitespace = false;
					}
				}
				headers.back().value.resize(s, '\0');
				s = 0;
				whitespace = true;
				for(i = k; i<j; ++i) {
					if(data[i]==' ' || data[i]=='\t') {
						if(!whitespace) {
							headers.back().value[s++] = ' ';
						}
						whitespace = true;
					} else {
						headers.back().value[s++] = data[i];
						whitespace = false;
					}
				}
				i = j+2;
			}
			s.EraseReadData(headerslen+4);
			state = HTTPREQUESTSTATE_WAIT_MESSAGEBODY;
		}
	}
	
	switch(s.GetState()) {
		case SOCKETSTATE_CONNECTED: {
			s.ShutDown();
			break;
		}
		case SOCKETSTATE_CLOSED: {
			state = HTTPREQUESTSTATE_CLOSED;
			break;
		}
		case SOCKETSTATE_ERROR: {
			s.Destroy();
			state = HTTPREQUESTSTATE_ERROR;
			break;
		}
	}
	
}

void HttpRequest::Reset() {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	if(s.GetState()!=SOCKETSTATE_NOTCONNECTED) {
		s.Destroy();
	}
	headers.clear();
	postparameters.clear();
	state = HTTPREQUESTSTATE_NOTCONNECTED;
}

unsigned int HttpRequest::GetStatusCode() {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	return statuscode;
}

unsigned int HttpRequest::GetHeaderCount() {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	return headers.size();
}
const char* HttpRequest::GetHeaderName(unsigned int i) {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	if(i>=headers.size()) return NULL;
	return headers[i].name.c_str();
}
const char* HttpRequest::GetHeaderValue(unsigned int i) {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	if(i>=headers.size()) return NULL;
	return headers[i].value.c_str();
}
const char* HttpRequest::GetHeader(const char* name) {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	for(unsigned int i = 0; i<headers.size(); ++i) {
		if(_strnicmp(headers[i].name.data(), name, headers[i].name.length())==0) {
			return headers[i].value.c_str();
		}
	}
	return NULL;
}

const char* HttpRequest::GetMessageBody() {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	if(state==HTTPREQUESTSTATE_WAIT_MESSAGEBODY || state==HTTPREQUESTSTATE_CLOSED) {
		return s.GetReadData();
	} else {
		return NULL;
	}
}

unsigned long HttpRequest::GetMessageBodyLength() {
	assert(state!=HTTPREQUESTSTATE_NOTCONNECTED);
	if(state==HTTPREQUESTSTATE_WAIT_MESSAGEBODY || state==HTTPREQUESTSTATE_CLOSED) {
		return s.GetReadDataLength();
	} else {
		return 0;
	}
}

std::string HttpRequest::UrlEncode(const char* str, bool keepspecialchars) {
	const char* hextable = "0123456789abcdef";
	unsigned char t = (keepspecialchars)? 1 : 2;
	unsigned long j = 0;
	for(const char *p = str; *p!='\0'; ++p) {
		if(httprequest_urlencodetable.table[(unsigned char)(*p)]>=t) {
			j += 1;
		} else {
			j += 3;
		}
	}
	std::string ret(j, '\0');
	j = 0;
	for(const char *p = str; *p!='\0'; ++p) {
		if(httprequest_urlencodetable.table[(unsigned char)(*p)]>=t) {
			ret[j++] = *p;
		} else {
			ret[j++] = '%';
			ret[j++] = hextable[(unsigned char)(*p)>>4];
			ret[j++] = hextable[(unsigned char)(*p)&15];
		}
	}
	return ret;
}

