#include "SOCKET.h"

#include <cassert>
#include <new>

#define SOCKET_BLOCKSIZE_POW2 12
#define SOCKET_BLOCKSIZE (1<<SOCKET_BLOCKSIZE_POW2)
#define SOCKET_BLOCKSIZE_MASK (~(SOCKET_BLOCKSIZE-1))
#define SOCKET_BLOCKSIZE_ROUNDUP(length) ((max(1, (length))+SOCKET_BLOCKSIZE-1)&SOCKET_BLOCKSIZE_MASK) // the size of the buffers can't be 0

// global

bool InitSockets() {
	static WSADATA wsadata;
	if(WSAStartup(MAKEWORD(2, 0), &wsadata)!=0) {
		return false;
	}
	if(LOBYTE(wsadata.wVersion)!=2 || HIBYTE(wsadata.wVersion)!=0) {
		WSACleanup();
		return false;
	}
	return true;
}

// ListeningSocket

ListeningSocket::ListeningSocket() {
	state = LISTENINGSOCKETSTATE_NOTLISTENING;
}

ListeningSocket::~ListeningSocket() {
	if(state!=LISTENINGSOCKETSTATE_NOTLISTENING) {
		Destroy();
	}
}

void ListeningSocket::Create(bool ipv6, const char* port) {
	assert(state==LISTENINGSOCKETSTATE_NOTLISTENING);
	addrinfo hints;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = (ipv6)? AF_INET6 : AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	hints.ai_flags = AI_PASSIVE;
	addrinfo *first_addrinfo = NULL;
	s = INVALID_SOCKET;
	int result = getaddrinfo(NULL, port, &hints, &first_addrinfo);
	if(result!=0 || first_addrinfo==NULL) {
		state = LISTENINGSOCKETSTATE_ERROR;
		return;
	}
	s = socket(first_addrinfo->ai_family, first_addrinfo->ai_socktype, first_addrinfo->ai_protocol);
	if(s==INVALID_SOCKET) {
		freeaddrinfo(first_addrinfo);
		state = LISTENINGSOCKETSTATE_ERROR;
		return;
	}
	result = bind(s, first_addrinfo->ai_addr, first_addrinfo->ai_addrlen);
	freeaddrinfo(first_addrinfo);
	if(result==SOCKET_ERROR) {
		closesocket(s);
		s = INVALID_SOCKET;
		state = LISTENINGSOCKETSTATE_ERROR;
		return;
	}
	result = listen(s, SOMAXCONN);
	if(result==SOCKET_ERROR) {
		closesocket(s);
		s = INVALID_SOCKET;
		state = LISTENINGSOCKETSTATE_ERROR;
		return;
	}
	{
		unsigned long a = 1;
		ioctlsocket(s, FIONBIO, &a);
	}
	state = LISTENINGSOCKETSTATE_LISTENING;
}

void ListeningSocket::Destroy() {
	assert(state!=LISTENINGSOCKETSTATE_NOTLISTENING);
	state = LISTENINGSOCKETSTATE_NOTLISTENING;
	if(s!=INVALID_SOCKET) {
		closesocket(s);
		s = INVALID_SOCKET;
	}
}

bool ListeningSocket::CanAccept() {
	assert(state!=LISTENINGSOCKETSTATE_NOTLISTENING);
	if(state==LISTENINGSOCKETSTATE_ERROR) return false;
	fd_set read;
	FD_ZERO(&read);
	FD_SET(s, &read);
	timeval t = {0, 0};
	int result = select(0, &read, NULL, NULL, &t);
	if(result==SOCKET_ERROR) {
		closesocket(s);
		s = INVALID_SOCKET;
		state = LISTENINGSOCKETSTATE_ERROR;
		return false;
	}
	return (FD_ISSET(s, &read))? true : false;
}

void ListeningSocket::Accept(Socket* socket) {
	assert(state!=LISTENINGSOCKETSTATE_NOTLISTENING && state!=LISTENINGSOCKETSTATE_ERROR);
	assert(socket->state==SOCKETSTATE_NOTCONNECTED);
	socket->Init();
	SOCKET s2 = accept(s, NULL, NULL);
	if(s2==INVALID_SOCKET) {
		socket->state = SOCKETSTATE_ERROR;
	} else {
		socket->s = s2;
	}
}

// Socket

void Socket::Init() {
	first_addrinfo = NULL;
	s = INVALID_SOCKET;
	readstart = 0;
	readlen = 0;
	readblocksize = SOCKET_BLOCKSIZE_ROUNDUP(SOCKET_BLOCKSIZE);
	writestart = 0;
	writelen = 0;
	writeblocksize = SOCKET_BLOCKSIZE_ROUNDUP(0);
	readbuffer = (char*)(malloc(readblocksize));
	if(readbuffer==NULL) {
		throw std::bad_alloc();
	}
	writebuffer = (char*)(malloc(writeblocksize));
	if(writebuffer==NULL) {
		free(readbuffer);
		throw std::bad_alloc();
	}
	state = SOCKETSTATE_CONNECTING;
}

void Socket::ResizeReadBuffer(unsigned long length) {
	unsigned long newblocksize = SOCKET_BLOCKSIZE_ROUNDUP(length);
	if(newblocksize!=readblocksize) {
		void *temp = realloc(readbuffer, newblocksize);
		if(temp==NULL) {
			throw std::bad_alloc();
		}
		readbuffer = (char*)(temp);
		readblocksize = newblocksize;
	}
}

void Socket::ResizeWriteBuffer(unsigned long length) {
	unsigned long newblocksize = SOCKET_BLOCKSIZE_ROUNDUP(length);
	if(newblocksize!=writeblocksize) {
		void *temp = realloc(writebuffer, newblocksize);
		if(temp==NULL) {
			throw std::bad_alloc();
		}
		writebuffer = (char*)(temp);
		writeblocksize = newblocksize;
	}
}

Socket::Socket() {
	state = SOCKETSTATE_NOTCONNECTED;
}

Socket::~Socket() {
	if(state!=SOCKETSTATE_NOTCONNECTED) {
		Destroy();
	}
}

void Socket::Create(const char* address, const char* port) {
	assert(state==SOCKETSTATE_NOTCONNECTED);
	Init();
	addrinfo hints;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	int result = getaddrinfo(address, port, &hints, &first_addrinfo);
	if(result!=0 || first_addrinfo==NULL) {
		state = SOCKETSTATE_ERROR;
		return;
	}
	for(current_addrinfo = first_addrinfo; current_addrinfo!=NULL; current_addrinfo = current_addrinfo->ai_next) {
		s = socket(current_addrinfo->ai_family, current_addrinfo->ai_socktype, current_addrinfo->ai_protocol);
		if(s==INVALID_SOCKET) {
			continue;
		}
		{
			unsigned long a = 1;
			ioctlsocket(s, FIONBIO, &a);
		}
		connect(s, current_addrinfo->ai_addr, current_addrinfo->ai_addrlen);
		// connect usually fails because this is a nonblocking socket
		return;
	}
	freeaddrinfo(first_addrinfo);
	first_addrinfo = NULL;
	state = SOCKETSTATE_ERROR;
}

void Socket::Destroy() {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	state = SOCKETSTATE_NOTCONNECTED;
	if(first_addrinfo!=NULL) {
		freeaddrinfo(first_addrinfo);
		first_addrinfo = NULL;
	}
	if(s!=INVALID_SOCKET) {
		closesocket(s);
		s = INVALID_SOCKET;
	}
	free(readbuffer);
	readbuffer = NULL;
	free(writebuffer);
	writebuffer = NULL;
}

bool Socket::Update() {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_ERROR) return true;
	if(state==SOCKETSTATE_CLOSED) return false;
	bool run, haschanged = false;
	do {
		run = false;
		fd_set read, write, except;
		FD_ZERO(&read);
		FD_SET(s, &read);
		FD_ZERO(&write);
		FD_SET(s, &write);
		FD_ZERO(&except);
		FD_SET(s, &except);
		timeval t = {0, 0};
		int result = select(0, &read, &write, &except, &t);
		if(result==SOCKET_ERROR) {
			if(state==SOCKETSTATE_CONNECTING) {
				freeaddrinfo(first_addrinfo);
				first_addrinfo = NULL;
				s = INVALID_SOCKET;
			} else {
				closesocket(s);
				s = INVALID_SOCKET;
			}
			state = SOCKETSTATE_ERROR;
			haschanged = true;
			break;
		}
		if(FD_ISSET(s, &except)) {
			if(state==SOCKETSTATE_CONNECTING) {
				// connection attempt has failed, try next address
				closesocket(s);
				for(current_addrinfo = current_addrinfo->ai_next; current_addrinfo!=NULL; current_addrinfo = current_addrinfo->ai_next) {
					s = socket(current_addrinfo->ai_family, current_addrinfo->ai_socktype, current_addrinfo->ai_protocol);
					if(s==INVALID_SOCKET) {
						continue;
					}
					{
						unsigned long a = 1;
						ioctlsocket(s, FIONBIO, &a);
					}
					connect(s, current_addrinfo->ai_addr, current_addrinfo->ai_addrlen);
					break;
				}
				if(current_addrinfo==NULL) {
					freeaddrinfo(first_addrinfo);
					first_addrinfo = NULL;
					s = INVALID_SOCKET;
					state = SOCKETSTATE_ERROR;
					haschanged = true;
				}
			} else {
				closesocket(s);
				s = INVALID_SOCKET;
				state = SOCKETSTATE_ERROR;
				haschanged = true;
			}
		} else {
			bool canread, canwrite;
			canread = (FD_ISSET(s, &read)!=0);
			canwrite = (FD_ISSET(s, &write)!=0);
			if(state==SOCKETSTATE_CONNECTING && canwrite) {
				// connection attempt has succeeded
				freeaddrinfo(first_addrinfo);
				first_addrinfo = NULL;
				state = SOCKETSTATE_CONNECTED;
				haschanged = true;
			}
			if(canwrite && writelen>writestart) {
				int result = send(s, writebuffer+writestart, writelen-writestart, 0);
				if(result==SOCKET_ERROR) {
					closesocket(s);
					s = INVALID_SOCKET;
					state = SOCKETSTATE_ERROR;
					haschanged = true;
				} else {
					if((unsigned long)(result)==writelen-writestart) run = true;
					writestart += (unsigned long)(result);
					if(writestart==writelen) {
						if(state==SOCKETSTATE_SHUTDOWN) {
							shutdown(s, SD_SEND);
						}
						writestart = 0;
						writelen = 0;
						ResizeWriteBuffer(writelen);
					} else if(writestart>=SOCKET_BLOCKSIZE) {
						writelen -= writestart;
						memmove(writebuffer, writebuffer+writestart, writelen);
						writestart = 0;
						ResizeWriteBuffer(writelen);
					}
				}
			}
			if(canread) {
				ResizeReadBuffer(readlen+SOCKET_BLOCKSIZE);
				int result = recv(s, readbuffer+readlen, readblocksize-readlen, 0);
				if(result==SOCKET_ERROR) {
					closesocket(s);
					s = INVALID_SOCKET;
					state = SOCKETSTATE_ERROR;
					haschanged = true;
				} else if(result==0) {
					// connection has been closed
					closesocket(s);
					s = INVALID_SOCKET;
					state = SOCKETSTATE_CLOSED;
					haschanged = true;
				} else {
					haschanged = true;
					if((unsigned long)(result)==readblocksize-readlen) {
						run = true;
					}
					readlen += result;
				}
			}
		}
	} while(run);
	return haschanged;
}

void Socket::ShutDown() {
	assert(state!=SOCKETSTATE_NOTCONNECTED && state!=SOCKETSTATE_CONNECTING);
	if(state==SOCKETSTATE_CONNECTED) state = SOCKETSTATE_SHUTDOWN;
}

const char* Socket::GetPeerAddress() {
	assert(state!=SOCKETSTATE_NOTCONNECTED && state!=SOCKETSTATE_CONNECTING);
	if(state==SOCKETSTATE_CLOSED || state==SOCKETSTATE_ERROR) return NULL;
	sockaddr addr;
	int size = sizeof(addr);
	if(getpeername(s, &addr, &size)==SOCKET_ERROR) {
		return NULL;
	}
	static char buffer[40];
	buffer[39] = '\0';
	char *ptr = buffer+39;
	if(addr.sa_family==AF_INET) {
		unsigned int i;
		i = ((sockaddr_in*)(&addr))->sin_addr.S_un.S_un_b.s_b4;
		do {
			*(--ptr) = '0'+i%10;
		} while((i /= 10)!=0);
		*(--ptr) = '.';
		i = ((sockaddr_in*)(&addr))->sin_addr.S_un.S_un_b.s_b3;
		do {
			*(--ptr) = '0'+i%10;
		} while((i /= 10)!=0);
		*(--ptr) = '.';
		i = ((sockaddr_in*)(&addr))->sin_addr.S_un.S_un_b.s_b2;
		do {
			*(--ptr) = '0'+i%10;
		} while((i /= 10)!=0);
		*(--ptr) = '.';
		i = ((sockaddr_in*)(&addr))->sin_addr.S_un.S_un_b.s_b1;
		do {
			*(--ptr) = '0'+i%10;
		} while((i /= 10)!=0);
		return ptr;
	}
	if(addr.sa_family==AF_INET6) {
		const char *hextable = "0123456789abcdef";
		for(unsigned int j = 8; j>0; ) {
			unsigned int i = ((sockaddr_in6*)(&addr))->sin6_addr._S6_un._S6_u16[--j];
			*(--ptr) = hextable[i&15];
			*(--ptr) = hextable[(i>>4)&15];
			*(--ptr) = hextable[(i>>8)&15];
			*(--ptr) = hextable[(i>>12)&15];
			if(j!=0) *(--ptr) = ':';
		}
		return ptr;
	}
	return NULL;
}

const char* Socket::GetReadData() {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_ERROR) return "";
	return readbuffer+readstart;
}

unsigned long Socket::GetReadDataLength() {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_ERROR) return 0;
	return readlen-readstart;
}

void Socket::EraseReadData(unsigned long length) {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_ERROR) return;
	if(length>readlen-readstart) length = readlen-readstart;
	readstart += length;
	if(readstart==readlen) {
		readstart = 0;
		readlen = 0;
		ResizeReadBuffer(readlen+SOCKET_BLOCKSIZE);
	} else if(readstart>=SOCKET_BLOCKSIZE) {
		readlen -= readstart;
		memmove(readbuffer, readbuffer+readstart, readlen);
		readstart = 0;
		ResizeReadBuffer(readlen+SOCKET_BLOCKSIZE);
	}
}

void Socket::Write(const char* data, unsigned long length) {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_SHUTDOWN || state==SOCKETSTATE_CLOSED || state==SOCKETSTATE_ERROR) return;
	char *buffer = Write(length);
	if(buffer!=NULL) memcpy(buffer, data, length);
}

char* Socket::Write(unsigned long length) {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_SHUTDOWN || state==SOCKETSTATE_CLOSED || state==SOCKETSTATE_ERROR) return NULL;
	if(length==0) return NULL;
	writelen += length;
	ResizeWriteBuffer(writelen);
	return writebuffer+writelen-length;
}

unsigned long Socket::GetWriteDataLength() {
	assert(state!=SOCKETSTATE_NOTCONNECTED);
	if(state==SOCKETSTATE_ERROR) return 0;
	return writelen-writestart;
}

